@@ -22,10 +22,12 @@ import subprocess
 import os
 import shutil
 import hadoopy._freeze
-import select
 import sys
 import json
 import tempfile
+import stat
+import multiprocessing
+import Queue
 
 # These two globals are only used in the follow function
 WARNED_HADOOP_HOME = False
@@ -77,9 +79,9 @@ def launch(in_name, out_name, script_path, partitioner=False, files=(), jobconfs
     :param out_name: Output path
     :param script_path: Path to the script (e.g., script.py)
     :param partitioner: If True, the partitioner is the value.
-    :param files: Extra files (other than the script) (string or list).  NOTE: Hadoop copies the files into working directory
-    :param jobconfs: Extra jobconf parameters (string or list)
-    :param cmdenvs: Extra cmdenv parameters (string or list)
+    :param files: Extra files (other than the script) (iterator).  NOTE: Hadoop copies the files into working directory
+    :param jobconfs: Extra jobconf parameters (iterator)
+    :param cmdenvs: Extra cmdenv parameters (iterator)
     :param copy_script: If True, the script is added to the files list.
     :param wait: If True, wait till the process is completed (default True) this is useful if you want to run multiple jobs concurrently by using the 'process' entry in the output.
     :param hstreaming: The full hadoop streaming path to call.
@@ -105,7 +107,10 @@ def launch(in_name, out_name, script_path, partitioner=False, files=(), jobconfs
     :returns: output: Iterator of (key, value) pairs
     :raises: subprocess.CalledProcessError: Hadoop error.
     :raises: OSError: Hadoop streaming not found.
+    :raises: TypeError: Input types are not correct.
     """
+    if isinstance(files, str) or isinstance(jobconfs, str) or isinstance(cmdenvs, str):
+        raise TypeError('files,  jobconfs, and cmdenvs must be iterators of strings and not strings!')
     out = {}
     try:
         hadoop_cmd = 'hadoop jar ' + hstreaming
@@ -159,8 +164,6 @@ def launch(in_name, out_name, script_path, partitioner=False, files=(), jobconfs
     if num_reducers:
         cmd += ['-numReduceTasks', "'%i'"%(int(num_reducers))]
     # Add files
-    if isinstance(files, str):
-        files = [files]
     if copy_script:
         files = list(files)
         files.append(script_path)
@@ -177,18 +180,14 @@ def launch(in_name, out_name, script_path, partitioner=False, files=(), jobconfs
     for f in files:
         cmd += ['-file', f]
     # Add jobconfs
-    if isinstance(jobconfs, str):
-        jobconfs = [jobconfs]
+    jobconfs = list(jobconfs)
     if name == None:
-        jobconfs = list(jobconfs)
         jobconfs.append('mapred.job.name=%s' % (job_name))
     else:
         jobconfs.append('mapred.job.name=%s' % (str(name)))
     for jobconf in jobconfs:
         cmd += ['-jobconf', jobconf]
     # Add cmdenv
-    if isinstance(cmdenvs, str):
-        cmdenvs = [cmdenvs]
     for cmdenv in cmdenvs:
         cmd += ['-cmdenv', cmdenv]
     # Add IO
@@ -240,9 +239,9 @@ def launch_frozen(in_name, out_name, script_path, frozen_tar_path=None,
     :param temp_path: HDFS path that we can use to store temporary files (default to _hadoopy_temp)
     :param partitioner: If True, the partitioner is the value.
     :param wait: If True, wait till the process is completed (default True) this is useful if you want to run multiple jobs concurrently by using the 'process' entry in the output.
-    :param files: Extra files (other than the script) (string or list).  NOTE: Hadoop copies the files into working directory
-    :param jobconfs: Extra jobconf parameters (string or list)
-    :param cmdenvs: Extra cmdenv parameters (string or list)
+    :param files: Extra files (other than the script) (iterator).  NOTE: Hadoop copies the files into working directory
+    :param jobconfs: Extra jobconf parameters (iterator)
+    :param cmdenvs: Extra cmdenv parameters (iterator)
     :param hstreaming: The full hadoop streaming path to call.
     :param name: Set the job name to this (default None, job name is the script name)
     :param use_typedbytes: If True (default), use typedbytes IO.
@@ -263,19 +262,21 @@ def launch_frozen(in_name, out_name, script_path, frozen_tar_path=None,
     :returns: output: Iterator of (key, value) pairs
     :raises: subprocess.CalledProcessError: Hadoop error.
     :raises: OSError: Hadoop streaming not found.
+    :raises: TypeError: Input types are not correct.
     """
+    if (('files' in kw and isinstance(kw['files'], str)) or
+        ('jobconfs' in kw and isinstance(kw['jobconfs'], str)) or
+        ('cmdenvs' in kw and isinstance(kw['cmdenvs'], str))):
+        raise TypeError('files,  jobconfs, and cmdenvs must be iterators of strings and not strings!')
     cmds = []
     if not frozen_tar_path:
-        freeze_out = hadoopy.freeze_script(script_path)
+        freeze_out = hadoopy.freeze_script(script_path, temp_path=temp_path)
         frozen_tar_path = freeze_out['frozen_tar_path']
         cmds += freeze_out['cmds']
     try:
-        jobconfs = kw['jobconfs']
+        jobconfs = list(kw['jobconfs'])
     except KeyError:
         jobconfs = []
-    else:
-        if isinstance(jobconfs, str):
-            jobconfs = [jobconfs]
     jobconfs.append('"mapred.cache.archives=%s#_frozen"' % frozen_tar_path)
     jobconfs.append('"mapreduce.job.cache.archives=%s#_frozen"' % frozen_tar_path)
     kw['copy_script'] = False
@@ -288,8 +289,30 @@ def launch_frozen(in_name, out_name, script_path, frozen_tar_path=None,
     return out
 
 
+def _local_reader(worker_queue_maxsize, q_recv, q_send, in_r_fd, in_w_fd, out_r_fd, out_w_fd):
+    os.close(in_r_fd)
+    os.close(in_w_fd)
+    os.close(out_w_fd)
+    q = Queue.Queue()
+    with hadoopy.TypedBytesFile(read_fd=out_r_fd) as tbfp_r:
+        for num, kv in enumerate(tbfp_r):
+            while True:
+                while not q_recv.poll() and not q.empty():
+                    q_send.send(q.get())
+                try:
+                    q.put_nowait(kv)
+                except Queue.Full:
+                    continue
+                else:
+                    break
+    while not q.empty():
+        q_send.send(q.get())
+    q_send.close()
+
+
 def launch_local(in_name, out_name, script_path, max_input=-1,
-                 files=(), cmdenvs=(), pipe=True, python_cmd='python', remove_tempdir=True, **kw):
+                 files=(), cmdenvs=(), pipe=True, python_cmd='python', remove_tempdir=True,
+                 worker_queue_maxsize=0, **kw):
     """A simple local emulation of hadoop
 
     This doesn't run hadoop and it doesn't support many advanced features, it
@@ -312,11 +335,12 @@ def launch_local(in_name, out_name, script_path, max_input=-1,
     :param out_name: Output path or None.  If None then output is not placed on HDFS, it is available through the 'output' key of the return value.
     :param script_path: Path to the script (e.g., script.py)
     :param max_input: Maximum number of Mapper inputs, if < 0 (default) then unlimited.
-    :param files: Extra files (other than the script) (string or list).  NOTE: Hadoop copies the files into working directory
-    :param cmdenvs: Extra cmdenv parameters (string or list)
+    :param files: Extra files (other than the script) (iterator).  NOTE: Hadoop copies the files into working directory
+    :param cmdenvs: Extra cmdenv parameters (iterator)
     :param pipe: If true (default) then call user code through a pipe to isolate it and stop bugs when printing to stdout.  See project docs.
     :param python_cmd: The python command to use. The default is "python".  Can be used to override the system default python, e.g. python_cmd = "python2.6"
     :param remove_tempdir: If True (default), then rmtree the temporary dir, else print its location.  Useful if you need to see temporary files or how input files are copied.
+    :param worker_queue_maxsize: The number of elements the queue holding results from the worker task will hold (default 0 which is unlimited).
     :rtype: Dictionary with some of the following entries (depending on options)
     :returns: freeze_cmds: Freeze command(s) ran
     :returns: frozen_tar_path: HDFS path to frozen file
@@ -325,27 +349,27 @@ def launch_local(in_name, out_name, script_path, max_input=-1,
     :returns: output: Iterator of (key, value) pairs
     :raises: subprocess.CalledProcessError: Hadoop error.
     :raises: OSError: Hadoop streaming not found.
+    :raises: TypeError: Input types are not correct.
     """
-    print('Hadoopy Local: In[%s] Out[%s] Script[%s]' % (in_name, out_name, script_path))
-    script_path = os.path.abspath(script_path)
+    if isinstance(files, str) or isinstance(cmdenvs, str) or ('cmdenvs' in kw and isinstance(kw['cmdenvs'], str)):
+        raise TypeError('files,  jobconfs, and cmdenvs must be iterators of strings and not strings!')
+    print('Local[%s]' % script_path)
     script_info = _parse_info(script_path, python_cmd)
-    if files:
-        if isinstance(files, str):
-            files = [files]
-        files = [os.path.abspath(f) for f in files]
+    if not files:
+        files = []
+    files.append(script_path)
+    files = [os.path.abspath(f) for f in files]
     # Setup env
     env = dict(os.environ)
     env['stream_map_input'] = 'typedbytes'
     if cmdenvs:
-        if isinstance(cmdenvs, str):
-            cmdenvs = [cmdenvs]
         for cmdenv in cmdenvs:
             k, v = cmdenv.split('=', 1)
             env[k] = v
 
     def _run_task(task, kvs, env):
-        print('Running [%s]' % task)
         sys.stdout.flush()
+        cur_max_input = max_input if task == 'map' else -1
         if pipe:
             task = 'pipe %s' % task
         in_r_fd, in_w_fd = os.pipe()
@@ -353,55 +377,87 @@ def launch_local(in_name, out_name, script_path, max_input=-1,
         cmd = ('%s %s %s' % (python_cmd, script_path, task)).split()
         a = os.fdopen(in_r_fd, 'r')
         b = os.fdopen(out_w_fd, 'w')
-        p = subprocess.Popen(cmd,
-                             stdin=a,
-                             stdout=b,
-                             close_fds=True,
-                             env=env)
-        a.close()
-        b.close()
-        with hadoopy.TypedBytesFile(read_fd=out_r_fd) as tbfp_r:
+        q_recv, q_send = multiprocessing.Pipe(True)
+        reader_process = multiprocessing.Process(target=_local_reader, args=(worker_queue_maxsize, q_recv, q_send, in_r_fd, in_w_fd, out_r_fd, out_w_fd))
+        reader_process.start()
+        os.close(out_r_fd)
+        q_send.close()
+        try:
+            p = subprocess.Popen(cmd,
+                                 stdin=a,
+                                 stdout=b,
+                                 close_fds=True,
+                                 env=env)
+            a.close()
+            b.close()
+            # main =(in_fd)> p
+            # main <(out_fd)= p
             with hadoopy.TypedBytesFile(write_fd=in_w_fd, flush_writes=True) as tbfp_w:
                 for num, kv in enumerate(kvs):
-                    if max_input >= 0 and max_input <= num:
+                    if cur_max_input >= 0 and cur_max_input <= num:
                         break
-                    # Use select to see if the buffer has anything in it
-                    # this minimizes the chance that the pipe will block
-                    while select.select([out_r_fd], [], [], 0)[0]:
-                        yield tbfp_r.next()
+                    if reader_process.exitcode:
+                        raise EOFError('Reader process died[%s]' % reader_process.exitcode)
+                    while q_recv.poll():
+                        try:
+                            yield q_recv.recv()
+                        except EOFError:
+                            break
                     tbfp_w.write(kv)
             # Get any remaining values
-            for kv in tbfp_r:
-                yield kv
-        p.wait()
+            while True:
+                if reader_process.exitcode:
+                    raise EOFError('Reader process died[%s]' % reader_process.exitcode)
+                try:
+                    yield q_recv.recv()
+                except EOFError:
+                    break
+            print('********  Done with inputs and cleaned up')
+        finally:
+            print('Closing up shop')
+            q_recv.close()
+            reader_process.join()
+            p.kill()
+            p.wait()
+
     orig_pwd = os.path.abspath('.')
     new_pwd = tempfile.mkdtemp()
     out = {}
     try:
+        print('Hadoopy: Launch local changing current directory to [%s]' % new_pwd)
         os.chdir(new_pwd)
         if files:
             for f in files:
                 shutil.copy(f, os.path.basename(f))
+        script_path = os.path.basename(script_path)
+        os.chmod(script_path, stat.S_IXUSR | stat.S_IRUSR | stat.S_IWUSR)
         if isinstance(in_name, str) or (in_name and isinstance(in_name, (list, tuple)) and isinstance(in_name[0], str)):
             in_kvs = hadoopy.readtb(in_name)
         else:
             in_kvs = in_name
-        kvs = list(_run_task('map', in_kvs, env))
         if 'reduce' in script_info['tasks']:
+            kvs = list(_run_task('map', in_kvs, env))
             if 'combine' in script_info['tasks']:
+                print('COMBINER ----------------------------')
                 kvs = hadoopy.Test.sort_kv(kvs)
                 kvs = list(_run_task('combine', kvs, env))
+            print('REDUCER ----------------------------')
             kvs = hadoopy.Test.sort_kv(kvs)
             kvs = _run_task('reduce', kvs, env)
+        else:
+            print('********** MapOnly Task')
+            kvs = _run_task('map', in_kvs, env)
     except OSError, e:
-        print('Error: Ensure that [%s] is executable and starts with "#!/usr/bin/env python".' % script_path)
+        print('Error: Ensure that [%s] starts with "#!/usr/bin/env python".' % script_path)
         raise e
     else:
         if out_name is not None:
+            print('Start writing')
             hadoopy.writetb(out_name, kvs)
+            print('Done writing')
             out['output'] = hadoopy.readtb(out_name)
         else:
-            out['output'] = kvs
+            out['output'] = iter(list(kvs))  # TODO(brandyn): Potential problem if using large values, fixes changing dir early
         return out
     finally:
         os.chdir(orig_pwd)
